-- Easy Shade 1.5
-- An automated lighting script for mapmapkers
-- By 'Crater Creator' with improvements by 'Hopper'

show_inspector = false;

function sunlight(...)
	local _contrast, _manual_sun_angle, _cumulative = ...;
	-- default values of parameters
	contrast = _contrast or 50;
	manual_sun_angle = _manual_sun_angle;
	cumulative = _cumulative;	-- default is false (nil)

	-- I like to give input in degrees
	if(manual_sun_angle) then
		manual_sun_angle = math.rad(manual_sun_angle);
	end
	
	-- undo the last shading operation
	if(not cumulative) then
		resetlights();
	end

    for S in Sides() do
		x1 = 0;
		x2 = 0;
		y1 = 0;
		y2 = 0;
		x1 = S.line.endpoints[0].x;
		y1 = S.line.endpoints[0].y;
		x2 = S.line.endpoints[1].x;
		y2 = S.line.endpoints[1].y;

		-- use atan2 to find angle properly
		-- http://en.wikipedia.org/wiki/Atan2#Motivation
		angle = math.atan2(y2-y1, x2-x1)
		-- angle = math.atan((y2-y1)/(x2-x1));

		-- do we need the opposite side instead?
		if S.line.counterclockwise_side == S then
			angle = angle + math.rad(180);
		end
		
		-- normalize to Marathon's angle system;
		-- this way, wall angle is the same as
		-- the player's direction when facing
		-- the wall head-on
		angle = angle - math.rad(90);
		-- if(x2<x1) then
		--     angle = math.rad(180) - angle;
		-- end

		-- confine range to [ 0 - 2pi )
		-- to make debugging easier
		local twopi = 2 * math.pi;
		while angle < 0 do
		  angle = angle + twopi;
		end
		while angle >= twopi do
		  angle = angle - twopi;
		end

		-- only use sun angle if one was passed in,
		-- otherwise use player facing angle
		if(manual_sun_angle) then
			sun_angle = manual_sun_angle;
		else
			sun_angle = math.rad(Players[0].yaw);
		end

		change = contrast / 5 * math.cos(angle-sun_angle);
		
		if(S.primary.light.index <= 20) then
			new_light_primary = BoundLight(S.primary.light.index - change);
			S.primary.light = Lights[new_light_primary];
		end
	
		if(S.secondary and S.secondary.light.index <= 20) then
			new_light_secondary = BoundLight(S.secondary.light.index - change);
			S.secondary.light = Lights[new_light_secondary];
		end

		if(S.transparent and S.transparent.light.index <= 20) then
			new_light_transparent = BoundLight(S.transparent.light.index - change);
			S.transparent.light = Lights[new_light_transparent];
		end
    end
end

-- helper functions
function BoundLight(light)
    -- don't go past light 0 or light 20; other lights are undefined
	light = math.min(light,20);
	light = math.max(light,0);
    -- light indices must be integers
	light = math.floor(light);
	return light;
end



function roomlight(...)
	local _brightness, _floor_threshold, _ceiling_threshold, _radius, _nudge = ...;
	-- default values of parameters
	brightness = _brightness or 50;
	floor_threshold = _floor_threshold or 0.5;
	ceiling_threshold = _ceiling_threshold or _floor_threshold or 0.5;
	radius = _radius or 64;
	nudge = _nudge;	-- default is false (nil)

	new_light = BoundLight(20 - brightness / 5);
	add_light = brightness / 5;
	sub_light = -brightness / 5;
	
	for p in Polygons() do
		p._checked = false;
	end
	still_checking = true;
	nPolysChecked = 0;
	nPolysAffected = 0;
	
	start_poly = Players[0].polygon;
	if(nudge == 1) then
		LightenPoly(start_poly, add_light);
	elseif(nudge == 2) then
		LightenPoly(start_poly, sub_light);
	else
		LightPoly(start_poly, new_light);
	end
	start_poly._checked = true;
	nPolysChecked = nPolysChecked + 1;
	nPolysAffected = nPolysAffected + 1;
	
	orig_x = Players[0].x;
	orig_y = Players[0].y;
	
	CheckAdjacentPolys(start_poly);
end

-- helper functions
function CheckAdjacentPolys(start_poly)
	for new_poly in start_poly:adjacent_polygons() do
		if(new_poly._checked == false) then
			new_poly._checked = true;
			nPolysChecked = nPolysChecked + 1;
			CheckTwoPolys(new_poly, start_poly);
		end
	end
end

function CheckTwoPolys(new_poly, old_poly)
	-- have we checked all polygons?
	if(nPolysChecked == (# Polygons)) then
		still_checking = false;
	end
	floor_change = math.abs(new_poly.floor.z - old_poly.floor.z);
	ceiling_change = math.abs(new_poly.ceiling.z - old_poly.ceiling.z);
	distance = math.sqrt((new_poly.x-orig_x)^2 + (new_poly.y-orig_y)^2);
	if(floor_change < floor_threshold and ceiling_change < ceiling_threshold and distance <= radius) then
		--Players.print("another poly will be affected.");
		if(nudge == 1) then
			LightenPoly(new_poly, add_light);
		elseif(nudge == 2) then
			LightenPoly(new_poly, sub_light);
		else
			LightPoly(new_poly, new_light);
		end
	
		if(still_checking) then
			CheckAdjacentPolys(new_poly);
		else
			Players.print(nPolysAffected.." polygons were changed.");
		end
	end
	
end

function LightPoly(polygon,light)
	if(polygon.floor.light.index <= 20) then
		nPolysAffected = nPolysAffected + 1;
		polygon.floor.light = light;
	end
	if(polygon.ceiling.light.index <= 20) then
		polygon.ceiling.light = light;
	end
	for s in polygon:sides() do
		if(s.primary.light.index <= 20) then
			s.primary.light = light;
		end
		if(s.secondary.light.index <= 20) then
			s.secondary.light =light;
		end
		if(s.transparent.light.index <= 20) then
			s.transparent.light = light;
		end
	end
end

function LightenPoly(polygon,lighter)
	if(polygon.floor.light.index <= 20) then
		nPolysAffected = nPolysAffected + 1;
		polygon.floor.light = BoundLight(polygon.floor.light.index - lighter);
	end
	if(polygon.ceiling.light.index <= 20) then
		polygon.ceiling.light = BoundLight(polygon.ceiling.light.index - lighter);
	end
	for s in polygon:sides() do
		if(s.primary.light.index <= 20) then
			s.primary.light = BoundLight(s.primary.light.index - lighter);
		end
		if(s.secondary.light.index <= 20) then
			s.secondary.light = BoundLight(s.secondary.light.index - lighter);
		end
		if(s.transparent.light.index <= 20) then
			s.transparent.light = BoundLight(s.transparent.light.index - lighter);
		end
	end
end



function pointlight(...)
	-- for now, algorithm is 2D
	local _brightness, _max_radius = ...;
	-- default values of parameters
	brightness = _brightness or 100;
	max_radius = _max_radius or 5;
	
	for P in Polygons() do
		distance = math.sqrt((P.x-Players[0].x)^2 + (P.y-Players[0].y)^2);
		if(distance <= max_radius) then
			-- for now, use a linear falloff
			add_intensity = (1 - distance/max_radius) * brightness;
			change = add_intensity / 5;
		
			new_light_floor = BoundLight(P.floor.light.index - change);
			P.floor.light = Lights[new_light_floor];

			new_light_ceiling = BoundLight(P.ceiling.light.index - change);
			P.floor.light = Lights[new_light_ceiling];
			for S in P:sides() do
				if(S.primary.light.index <= 20) then
					new_light_primary = BoundLight(S.primary.light.index - change);
					S.primary.light = Lights[new_light_primary];
				end
			
				if(S.secondary and S.secondary.light.index <= 20) then
					new_light_secondary = BoundLight(S.secondary.light.index - change);
					S.secondary.light = Lights[new_light_secondary];
				end
		
				if(S.transparent and S.transparent.light.index <= 20) then
					new_light_transparent = BoundLight(S.transparent.light.index - change);
					S.transparent.light = Lights[new_light_transparent];
				end
			end
		end
		
	end
end


-- syntactic sugar functions
function lighter(...)
	local intensity = ...;
	-- default values of parameters
	intensity = _intensity or 5;
	roomlight(intensity, 0.5, 0.5, 64, 1);
end

function darker(...)
	local intensity = ...;
	-- default values of parameters
	intensity = _intensity or 5;
	roomlight(intensity, 0.5, 0.5, 64, 2);
end

function pavelight(...)
	local _brightness = ...;
	-- default values of parameters
	brightness = _brightness or 50;
	roomlight(brightness,18,18,64);	
end

function inspect()
	show_inspector = not show_inspector;
end

function StoreShades()
	for S in Sides() do
		if S.primary then
			S.primary._orig_light = S.primary.light
		end
		if S.secondary then
			S.secondary._orig_light = S.secondary.light
		end
		if S.transparent then
			S.transparent._orig_light = S.transparent.light
		end
	end
	for P in Polygons() do
		P._orig_floor_light = P.floor.light;
		P._orig_ceiling_light = P.ceiling.light;
	end
end

function resetlights()
	for S in Sides() do
		if S.primary then
			S.primary.light = S.primary._orig_light
		end
		if S.secondary then
			S.secondary.light = S.secondary._orig_light
		end
		if S.transparent then
			S.transparent.light = S.transparent._orig_light
		end
	end
	for P in Polygons() do
		P.floor.light = P._orig_floor_light;
		P.ceiling.light = P._orig_ceiling_light;
	end
end

SShades = {}
function SShades.init()
	StoreShades();
	SShades.shaded = false;
end
function SShades.toggle()
	if SShades.shaded then
		resetlights();
		SShades.shaded = false;
		Players.print("Auto shading off");
	else
		sunlight(contrast);
		SShades.shaded = true;
		Players.print("Auto shading on");
	end
end
function SShades.update()
	for p in Players() do
		if(p._keys.mic_prev and (not p._keys.mic)) then
			SShades.toggle();
			return;
		elseif(p._keys.prev_weap) then
			darker();
			return;
		elseif(p._keys.next_weap) then
			lighter();
			return;
		end
	end
end
		

SKeys = {}
function SKeys.init()
	for p in Players() do
		p._keys = {};
		p._keys.mic = false;
		p._keys.mic_prev = false;
		p._keys.prev_weap = false;
		p._keys.next_weap = false;
		p._keys.prev_weap = false;
	end
end
function SKeys.update()
	for p in Players() do
		p._keys.mic_prev = p._keys.mic;
		p._keys.mic = p.action_flags.microphone_button;
		p.action_flags.microphone_button = false;

		p._keys.prev_weap_prev = p._keys.prev_weap;
		p._keys.prev_weap = p.action_flags.cycle_weapons_backward;
		--p.action_flags.cycle_weapons_backward = false;

		p._keys.next_weap_prev = p._keys.next_weap;
		p._keys.next_weap = p.action_flags.cycle_weapons_forward;
		--p.action_flags.cycle_weapons_backward = false;
	end
end

SInspector = {}
function SInspector.init()
	-- no need
end
function SInspector.update()
	if ( (Game.ticks % 6) == 0 and show_inspector ) then
		for p in Players() do
			local msg = ""
			local S,x,y,z,poly = p:find_target()
			if S and S.line then
		x1 = 0;
		x2 = 0;
		y1 = 0;
		y2 = 0;
		--if(S.line.endpoints[0].index < S.line.endpoints[1].index) then
			x1 = S.line.endpoints[0].x;
			y1 = S.line.endpoints[0].y;
			x2 = S.line.endpoints[1].x;
			y2 = S.line.endpoints[1].y;
		--else
		--	x1 = S.line.endpoints[1].x;
		--	y1 = S.line.endpoints[1].y;
		--	x2 = S.line.endpoints[0].x;
		--	y2 = S.line.endpoints[0].y;
		--end
				
				local angle = math.atan((y2-y1)/(x2-x1))
		if S.line.counterclockwise_side == S then
			angle = angle + math.rad(180)
		end
				if(x2<x1) then
						angle = math.rad(180) - angle
				end
				msg = msg .. "Angle: " .. math.floor(math.deg(angle)) .. "	"
				--msg = msg .. "x2: " .. x2 .. "	"

				if S.primary then
					msg = msg .. "Pri: " .. S.primary.light.index .. "	"
				end
				if S.secondary then
					msg = msg .. "Sec: " .. S.secondary.light.index .. "	"
				end
				if S.transparent then
					msg = msg .. "Trans: " .. S.transparent.light.index .. "	"
				end
			end
			Players.print(msg)
		end
	end
end

function init()
	Players.print("Script EasyShade 1.5 loaded.");
	SKeys.init();
	SInspector.init();
	SShades.init();
	for p in Polygons() do
		p._checked = false;
	end
	
	inited_script = true;
end

Triggers = {}
function Triggers.idle()
	if not inited_script then init() end
	
	SKeys.update()
	SInspector.update()
	SShades.update()
end

